package config

import (
    "os"
    "sync"
    "strings"
    "path/filepath"
    "nexus-server/logger"
    "gopkg.in/ini.v1"
    "github.com/sirupsen/logrus"
)

type Config struct {
    Core struct {
        TmpDir string `ini:"tmp_dir"`
        LogLevel string `ini:"log_level"`
        APIDefaultPort int `ini:"api_default_port"`
        VMSpiceMinPort int `ini:"vm_spice_min_port"`
        VMSpiceMaxPort int `ini:"vm_spice_max_port"`
    } `ini:"core"`

    Auth struct {
        AccessTokenExpirationTimeInHours int `ini:"access_token_expiration_time_hours"`
        MaxLoginAttempts int `ini:"max_login_attempts"`
        LockedTimeInHours float64 `ini:"locked_time_hours"`
    } `ini:"auth"`

    Limits struct {
        MaxUploadSize string `ini:"max_upload_size"`
        KsmRamSaving float64 `ini:"ksm_ram_saving"`
        RamUsageLimit float64 `ini:"ram_usage_limit"`
    } `ini:"limits"`

    Tools struct {
        QemuSystem string `ini:"qemu_system"`
        QemuImg string `ini:"qemu_img"`
        Guestfish string `ini:"guestfish"`
    } `ini:"tools"`

    // Unix permissions when creating directories: when creating templates and VMs and for tmp directory.
    MkDirPerm os.FileMode

    VMSpicePwdLength int
    VMSpicePwdDigitCount int
    VMSpicePwdSymbolCount int
    VMSpicePwdRepeatChars bool

    VMAttachPwdLength int
    VMAttachPwdDigitCount int
    VMAttachPwdSymbolCount int
    VMAttachPwdRepeatChars bool

    UsersFile string
    DataDir string
    VMsDir string
    TemplatesDir string
    CertsDir string
}

var log = logger.GetInstance()
var config *Config
var once sync.Once

func GetInstance() *Config {
    once.Do(func() {
        certsDir := "../certs"
        configDir := "../config"
        dataDir := "../data"
    
        conf, err := ini.LoadSources(ini.LoadOptions{
            Loose: false,
        }, filepath.Join(configDir, "nexus.conf"))
        if err != nil {
            log.Fatal("Failed reading config file: "+err.Error())
        }
    
        config = &Config {}
        err = conf.StrictMapTo(config)
        if err != nil {
           log.Fatal("Failed parsing config file: "+err.Error())
        }
    
        sanityChecks(config)

        config.MkDirPerm = 0750

        // Spice password
        config.VMSpicePwdLength = 32
        config.VMSpicePwdDigitCount = 10
        config.VMSpicePwdSymbolCount = 0
        config.VMSpicePwdRepeatChars = true

        // VM attach password
        config.VMAttachPwdLength = 10
        config.VMAttachPwdDigitCount = 4
        config.VMAttachPwdSymbolCount = 0
        config.VMAttachPwdRepeatChars = true
        
        config.UsersFile = filepath.Join(configDir, "/users.json")
        config.DataDir = dataDir
        config.VMsDir = filepath.Join(dataDir, "/vms")
        config.TemplatesDir = filepath.Join(dataDir, "/templates")
        config.CertsDir = certsDir
    
        // // Deletes temp directory and its content.
        // if err := os.RemoveAll(config.Core.TmpDir); err != nil {
        //     log.Fatal("Failed deleting tmp directory: "+err.Error())
        // }
    
        // if err := os.Mkdir(config.Core.TmpDir, config.MkDirPerm); err != nil {
        //     log.Fatal("Failed creating tmp dir: "+err.Error())
        // }
    })
    return config
}

func sanityChecks(c *Config) {
    checkDefinedAndExists(c.Core.TmpDir, "tmp_dir")

    if c.Core.LogLevel == "" {
        log.Fatal("Config file error: missing \"log_level\" value")
    }
    loglevelStr := strings.ToLower(c.Core.LogLevel)
    switch loglevelStr {
        case "debug":
            log.SetLevel(logrus.DebugLevel)
        case "info":
            log.SetLevel(logrus.InfoLevel)
        case "warn":
            log.SetLevel(logrus.WarnLevel)
        case "error":
            log.SetLevel(logrus.ErrorLevel)
        case "fatal":
            log.SetLevel(logrus.FatalLevel)
        default:
            log.Fatal("Config file error: invalid \"log_level\" value")
    }

    if c.Core.APIDefaultPort < 1025 || c.Core.APIDefaultPort > 65535 {
        log.Fatal("Config file error: invalid or missing \"api_default_port\" value")
    }
    if c.Core.VMSpiceMinPort < 1025 || c.Core.VMSpiceMinPort > 65535 {
        log.Fatal("Config file error: invalid or missing \"vm_spice_min_port\" value")
    }
    if c.Core.VMSpiceMaxPort < 1025 || c.Core.VMSpiceMaxPort > 65535 {
        log.Fatal("Config file error: invalid or missing \"vm_spice_max_port\" value")
    }
    if c.Core.VMSpiceMaxPort <= c.Core.VMSpiceMinPort {
        log.Fatal("Config file error: \"vm_spice_max_port\" must be > \"vm_spice_min_port\"")
    }

    if c.Auth.AccessTokenExpirationTimeInHours <= 0 {
        log.Fatal("Config file error: invalid or missing \"access_token_expiration_time_hours\" value")
    }
    if c.Auth.MaxLoginAttempts < 1 {
        log.Fatal("Config file error: invalid or missing \"max_login_attempts\" value")
    }
    if c.Auth.LockedTimeInHours <= 0.0 {
        log.Fatal("Config file error: invalid or missing \"locked_time_hours\" value")
    }

    if c.Limits.MaxUploadSize == "" {
        log.Fatal("Config file error: missing \"max_upload_size\" value")
    }
    if c.Limits.KsmRamSaving < 0.0 || c.Limits.KsmRamSaving >= 1.0 {
        log.Fatal("Config file error: invalid \"ksm_ram_saving\" value")
    }
    if c.Limits.RamUsageLimit < 0.0 || c.Limits.RamUsageLimit >= 1.0 {
        log.Fatal("Config file error: invalid \"ram_usage_limit\" value")
    }

    checkDefinedAndExists(c.Tools.QemuSystem, "qemu_system")
    checkDefinedAndExists(c.Tools.QemuImg, "qemu_img")
    checkDefinedAndExists(c.Tools.Guestfish, "guestfish")
}

func checkDefinedAndExists(filePath, propertyName string) {
    if filePath == "" {
        log.Fatal("Config file error: missing \"tmp_dir\" value")
    }
    _, err := os.Stat(filePath)
    if err != nil {
        log.Fatal("Config file error: \""+propertyName+"\": \""+filePath+"\": "+err.Error())
    }
}
